import logging
from copy import deepcopy
from typing import Any, Dict

import torch

from rlearn.sports.soccer.dataclass import (
    SimpleObservation_PVS,
    SimpleObservation_EDMS,
)
from rlearn.sports.soccer.constant import ONBALL_ACTION_INDICES
from rlearn.sports.soccer.modules.datamodule.datamodule import DataModule
from rlearn.sports.soccer.modules.state_action_tokenizer.state_action_tokenizer import StateActionTokenizerBase

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


@DataModule.register("jleague_rl_attacker")
class JLeagueRLAttackerDataModule(DataModule):
    def __init__(
        self,
        state_action_tokenizer: Dict[str, Any],
        train_dataset: torch.utils.data.Dataset,
        valid_dataset: torch.utils.data.Dataset | None = None,
        test_dataset: torch.utils.data.Dataset | None = None,
        batch_size: int = 128,
        num_workers: int = 8,
    ) -> None:
        super().__init__()
        self.train_dataset = train_dataset
        self.valid_dataset = valid_dataset
        self.test_dataset = test_dataset
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.state_action_tokenizer = StateActionTokenizerBase.from_params(state_action_tokenizer)

        self.onball_action = ONBALL_ACTION_INDICES

    @classmethod
    def preprocess_data(
        cls,
        dataset: torch.utils.data.Dataset,
        state_def: str,
        state_action_tokenizer: Dict[str, Any],
        num_workers: int = 8,
        preprocess_batch_size: int = 32,
    ) -> torch.utils.data.Dataset:
        state_action_tokenizer = StateActionTokenizerBase.from_params(deepcopy(state_action_tokenizer))
        dataset = dataset.map(
            cls._preprocess_data,
            batched=True,
            num_proc=num_workers,
            batch_size=preprocess_batch_size,
            fn_kwargs={"state_action_tokenizer": state_action_tokenizer, "state_def": state_def},
        )
        return dataset

    @classmethod
    def _preprocess_data(
        cls, examples, state_action_tokenizer: StateActionTokenizerBase, state_def: str
    ) -> Dict[str, torch.Tensor]:
        observation_in_examples = []
        action_in_examples = []
        reward_in_examples = []
        for sequence in examples["sequence"]:
            observation_in_example = []  # (seq_len, obs_dim)
            action_in_example = []  # (seq_len,)
            reward_in_example = []  # (seq_len,)
            for observation_action in sequence:
                observation = observation_action["observation"]
                action = observation_action["action"]
                reward = observation_action["reward"]
                assert isinstance(observation, dict), f"observation is not dict: {observation}"
                assert isinstance(action, str), f"action is not str: {action}"
                assert isinstance(reward, float), f"reward is not float: {reward}"
                if state_def == "PVS":
                    observation_in_example.append(SimpleObservation_PVS.from_dict(observation).to_tensor())
                elif state_def == "EDMS":
                    direction = 1
                    observation_in_example.append(SimpleObservation_EDMS.from_dict(observation).to_tensor(direction))
                if action in {"ball_recovery", "interception", "clearance", "pressure", "block"}:
                    action = "defensive_action"
                elif action == "goal":
                    action = "shot"
                action_in_example.append(state_action_tokenizer.encode(action))
                reward_in_example.append(reward)
            observation_in_examples.append(torch.stack(observation_in_example, dim=0))
            action_in_examples.append(torch.tensor(action_in_example, dtype=torch.long))
            reward_in_examples.append(torch.tensor(reward_in_example, dtype=torch.float32))
        return {
            "observation": observation_in_examples,  # List of np.array with (seq_len, obs_dim)
            "action": action_in_examples,  # List of np.array with (seq_len,)
            "reward": reward_in_examples,  # List of np.array with (seq_len,)
        }

    def train_dataloader(
        self,
        batch_size: int | None = None,
        shuffle: bool = True,
        num_workers: int | None = None,
    ) -> torch.utils.data.DataLoader:
        return self.build_dataloader(
            dataset=self.train_dataset,
            batch_size=batch_size or self.batch_size,
            shuffle=shuffle,
            num_workers=num_workers or self.num_workers,
        )

    def val_dataloader(
        self,
        batch_size: int | None = None,
        shuffle: bool = False,
        num_workers: int | None = None,
    ) -> torch.utils.data.DataLoader:
        assert self.valid_dataset is not None, f"valid dataset is not found: {self.valid_dataset}"
        return self.build_dataloader(
            dataset=self.valid_dataset,
            batch_size=batch_size or self.batch_size,
            shuffle=shuffle,
            num_workers=num_workers or self.num_workers,
        )

    def test_dataloader(
        self,
        batch_size: int | None = None,
        shuffle: bool = False,
        num_workers: int | None = None,
    ) -> torch.utils.data.DataLoader:
        assert self.test_dataset is not None, f"test dataset is not found: {self.test_dataset}"
        return self.build_dataloader(
            dataset=self.test_dataset,
            batch_size=batch_size or self.batch_size,
            shuffle=shuffle,
            num_workers=num_workers or self.num_workers,
        )

    def build_dataloader(
        self,
        dataset: torch.utils.data.Dataset,
        batch_size: int | None = None,
        shuffle: bool = False,
        num_workers: int = 0,
    ) -> torch.utils.data.DataLoader:
        return torch.utils.data.DataLoader(
            dataset=dataset,
            batch_size=batch_size or self.batch_size,
            shuffle=shuffle,
            collate_fn=self.batch_collator,
            num_workers=self.num_workers,
        )

    def batch_collator(self, instances) -> Dict[str, torch.Tensor]:
        max_length = max([len(instance["observation"]) for instance in instances])
        observation_dim = len(instances[0]["observation"][0])
        observation = torch.zeros((len(instances), max_length, observation_dim), dtype=torch.float32)
        action = torch.full(
            (len(instances), max_length),
            self.state_action_tokenizer.encode("[PAD]"),
            dtype=torch.long,
        )
        reward = torch.full(
            (len(instances), max_length),
            0,
            dtype=torch.float32,
        )
        mask = torch.zeros((len(instances), max_length), dtype=torch.bool)
        onball_mask = torch.zeros((len(instances), max_length), dtype=torch.bool)

        for i, instance in enumerate(instances):
            length = len(instance["observation"])
            observation[i, :length] = torch.tensor(instance["observation"], dtype=torch.float32)
            action[i, :length] = torch.tensor(instance["action"], dtype=torch.long)
            reward[i, :length] = torch.tensor(instance["reward"], dtype=torch.float32)
            mask[i, :length] = 1
            for j in range(length):
                action_idx = instance["action"][j]
                onball_mask[i, j] = 1 if action_idx in self.onball_action else 0

        batch = {
            "observation": observation,
            "action": action,
            "reward": reward,
            "mask": mask,
            "onball_mask": onball_mask,
        }

        return batch



import torch
import torch.nn as nn

from rlearn.sports.soccer.modules.token_embedder.token_embedder import TokenEmbedder


@TokenEmbedder.register("embedding")
class Embedding(TokenEmbedder):
    def __init__(
        self,
        embedding_dim: int,
        num_embeddings: int,
        padding_idx: int | None = None,
        max_norm: float | None = None,
        norm_type: float = 2.0,
        scale_grad_by_freq: bool = False,
        sparse: bool = False,
        _weight: torch.FloatTensor | None = None,
    ) -> None:
        super().__init__()
        self.embedding = nn.Embedding(
            num_embeddings=num_embeddings,
            embedding_dim=embedding_dim,
            padding_idx=padding_idx,
            max_norm=max_norm,
            norm_type=norm_type,
            scale_grad_by_freq=False,
            sparse=sparse,
            _weight=_weight,
        )
        self.padding_idx = padding_idx
        self.output_dim = embedding_dim
        self.input_dim = num_embeddings

    def get_output_dim(self) -> int:
        return self.output_dim

    def get_input_dim(self) -> int:
        return self.input_dim

    def forward(self, tokens: torch.Tensor) -> torch.Tensor:
        embedded = self.embedding(tokens)
        return embedded  # type: ignore



from typing import Dict, List

from tango.common import Registrable

from rlearn.sports.soccer.constant import FIELD_LENGTH, FIELD_WIDTH
from rlearn.sports.soccer.dataclass import Position, Velocity
from rlearn.sports.soccer.modules.state_action_tokenizer.preprocess_frames import discretize_direction


class StateActionTokenizerBase(Registrable):
    def __init__(
        self, action2id: Dict[str, int], position_granularity: float = 0.5, origin_pos: str = "center"
    ) -> None:
        self.action2id = action2id
        self.position_granularity = position_granularity  # unit: m
        self.origin_pos = origin_pos
        self.num_actions = len(action2id)

    def encode(self, token: str | Position | Velocity) -> int:
        raise NotImplementedError

    def decode(self, token_id: int) -> str | Position:
        raise NotImplementedError

    def _action2id(self, action: str) -> int:
        return self.action2id[action]

    def _velocity2action(self, velocity: Velocity) -> str:
        raise NotImplementedError

    def _position2id(self, position: Position) -> int:
        raise NotImplementedError

    @property
    def encode_position_separately(self) -> bool:
        raise NotImplementedError


@StateActionTokenizerBase.register("simple")
class SimpleStateActionTokenizer(StateActionTokenizerBase):
    def __init__(
        self,
        action2id: Dict[str, int],
        position_granularity: float = 0.5,
        origin_pos: str = "center",
        special_tokens: List[str] = ["[SEP]", "[PAD]", "[ACTION_SEP]"],
    ) -> None:
        self.action2id = action2id
        self.id2action = {id_: action for action, id_ in action2id.items()}
        self.position_granularity = position_granularity  # unit: m
        self.origin_pos = origin_pos
        self.num_actions = len(action2id)
        self.num_positions = int(2 * FIELD_LENGTH / position_granularity) * int(2 * FIELD_WIDTH / position_granularity)
        self.num_normal_tokens = self.num_actions + self.num_positions
        self.special_tokens = special_tokens
        self.num_tokens = self.num_normal_tokens + len(self.special_tokens)

    def encode_position_separately(self) -> bool:
        return False

    def encode(self, token: str | Position | Velocity) -> int:
        # if str, then token is action -> return action_id
        if isinstance(token, str):
            if token in self.special_tokens:
                return self.num_normal_tokens + self.special_tokens.index(token)
            else:
                return self._action2id(token)
        elif isinstance(token, Position):
            return self._position2id(token)
        elif isinstance(token, Velocity):
            return self._action2id(self._velocity2action(token))
        else:
            raise TypeError(f"token must be str or Position, but got {type(token)}")

    def decode(self, token_id: int) -> str | Position:
        if token_id < self.num_actions:
            return self._id2action(token_id)
        elif token_id < self.num_normal_tokens:
            return self._id2position(token_id)
        elif token_id < self.num_normal_tokens + len(self.special_tokens):
            return self.special_tokens[token_id - self.num_normal_tokens]
        else:
            raise ValueError(f"token_id must be less than {self.num_normal_tokens + len(self.special_tokens)}")

    def _action2id(self, action: str) -> int:
        return self.action2id[action]

    def _position2id(self, position: Position) -> int:
        # position.x is [-FIELD_LENGTH, FIELD_LENGTH] -> discretize to 0.5m (note that this is relative coordinate)
        # position.y is [-FIELD_WIDTH, FIELD_WIDTH] -> discretize to 0.5m (note that this is relative coordinate)
        x = int(
            (position.x + FIELD_LENGTH) / self.position_granularity
        )  # [0, 2 * FIELD_LENGTH / position_granularity]
        y = int((position.y + FIELD_WIDTH) / self.position_granularity)  #  [0, 2 * FIELD_WIDTH / position_granularity]
        assert 0 <= x <= int(2 * FIELD_LENGTH / self.position_granularity), f"position.x: {position.x}, x: {x}"
        assert 0 <= y <= int(2 * FIELD_WIDTH / self.position_granularity), f"position.y: {position.y}, y: {y}"
        return (
            x * int(2 * FIELD_WIDTH / self.position_granularity) + y + self.num_actions
        )  # [0, 2 * FIELD_LENGTH / position_granularity * 2 * FIELD_WIDTH / position_granularity + num_actions]

    def _velocity2action(self, velocity: Velocity) -> str:
        return discretize_direction(velocity.x, velocity.y)


@StateActionTokenizerBase.register("xy_separate")
class XYSeparateTokenizer(StateActionTokenizerBase):
    def __init__(
        self,
        action2id: Dict[str, int],
        position_granularity: float = 0.5,
        origin_pos: str = "center",
        special_tokens: List[str] = ["[SEP]", "[PAD]", "[ACTION_SEP]"],
    ) -> None:
        self.action2id = action2id
        self.id2action = {id_: action for action, id_ in action2id.items()}
        self.position_granularity = position_granularity  # unit: m
        self.origin_pos = origin_pos
        self.num_actions = len(action2id)
        self.num_positions = int(2 * FIELD_LENGTH / position_granularity)
        self.num_normal_tokens = self.num_actions + self.num_positions
        self.special_tokens = special_tokens
        self.num_tokens = self.num_normal_tokens + len(self.special_tokens)

    def encode_position_separately(self) -> bool:
        return True

    def encode(self, token: str | float | Velocity) -> int:
        # if str, then token is action -> return action_id
        if isinstance(token, str):
            if token in self.special_tokens:
                return self.num_normal_tokens + self.special_tokens.index(token)
            else:
                return self._action2id(token)
        elif isinstance(token, float):
            return self._coordinate2id(token)
        elif isinstance(token, Velocity):
            return self._action2id(self._velocity2action(token))
        else:
            raise TypeError(f"token must be str, float or Velocity, but got {type(token)}")

    def decode(self, token_id: int) -> str | Position:
        if token_id < self.num_actions:
            return self._id2action(token_id)
        elif token_id < self.num_normal_tokens:
            return self._id2position(token_id)
        elif token_id < self.num_normal_tokens + len(self.special_tokens):
            return self.special_tokens[token_id - self.num_normal_tokens]
        else:
            raise ValueError(f"token_id must be less than {self.num_normal_tokens + len(self.special_tokens)}")

    def _action2id(self, action: str) -> int:
        return self.action2id[action]

    def _id2action(self, id_: int) -> str:
        return self.id2action[id_]

    def _coordinate2id(self, coordinate: float) -> int:
        # range is [-FIELD_LENGTH, FIELD_LENGTH]
        id_ = int(
            (int(coordinate) + FIELD_LENGTH) / self.position_granularity
        )  # [0, 2 * FIELD_LENGTH / position_granularity]
        assert 0 <= id_ <= self.num_positions, f"coordinate: {coordinate}, id: {id_}"
        return id_ + self.num_actions

    def _velocity2action(self, velocity: Velocity) -> str:
        return discretize_direction(velocity.x, velocity.y)


@StateActionTokenizerBase.register("action_only")
class ActionOnlyTokenizer(StateActionTokenizerBase):
    def __init__(
        self,
        action2id: Dict[str, int],
        origin_pos: str = "center",
        special_tokens: List[str] = ["[SEP]", "[PAD]", "[ACTION_SEP]"],
    ) -> None:
        super().__init__(action2id, position_granularity=0.5, origin_pos=origin_pos)
        self.action2id = action2id
        self.id2action = {id_: action for action, id_ in action2id.items()}
        self.origin_pos = origin_pos
        self.num_actions = len(action2id)
        self.num_normal_tokens = self.num_actions
        self.special_tokens = special_tokens
        self.num_tokens = self.num_normal_tokens + len(self.special_tokens)

    def encode_position_separately(self) -> bool:
        return False

    def encode(self, token: str) -> int:
        # if str, then token is action -> return action_id
        if isinstance(token, str):
            if token in self.special_tokens:
                return self.num_normal_tokens + self.special_tokens.index(token)
            else:
                return self._action2id(token)
        else:
            raise TypeError(f"token must be str, but got {type(token)}")

    def decode(self, token_id: int) -> str | Position:
        if token_id < self.num_actions:
            return self._id2action(token_id)
        elif token_id < self.num_normal_tokens + len(self.special_tokens):
            return self.special_tokens[token_id - self.num_normal_tokens]
        else:
            raise ValueError(f"token_id must be less than {self.num_normal_tokens + len(self.special_tokens)}")

    def _action2id(self, action: str) -> int:
        return self.action2id[action]



from typing import cast

import torch
from torch import nn

from rlearn.sports.soccer.modules.seq2seq_encoder.seq2seq_encoder import Seq2SeqEncoder


class PytorchSeq2SeqWrapper(Seq2SeqEncoder):
    def __init__(self, module: torch.nn.Module) -> None:
        try:
            if not module.batch_first:
                raise ValueError("PytorchSeq2SeqWrapper only supports batch_first=True")
        except AttributeError:
            pass
        super().__init__()
        self._module = module
        try:
            self._is_bidirectional = cast(bool, self._module.bidirectional)
        except AttributeError:
            self._is_bidirectional = False
        if self._is_bidirectional:
            self._num_directions = 2
        else:
            self._num_directions = 1

    def get_input_dim(self) -> int:
        return cast(int, self._module.input_size)

    def get_output_dim(self) -> int:
        return cast(int, self._module.hidden_size) * self._num_directions

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        output, _ = self._module(inputs)
        return cast(torch.Tensor, output)


@Seq2SeqEncoder.register("gru")
class GruSeq2SeqEncoder(PytorchSeq2SeqWrapper):
    """
    Registered as a `Seq2SeqEncoder` with name "gru".
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        num_layers: int = 1,
        bias: bool = True,
        dropout: float = 0.0,
        bidirectional: bool = False,
    ):
        module = torch.nn.GRU(  # type: ignore
            input_size=input_size,
            hidden_size=hidden_size,
            num_layers=num_layers,
            bias=bias,
            batch_first=True,
            dropout=dropout,
            bidirectional=bidirectional,
        )
        super().__init__(module=module)


@Seq2SeqEncoder.register("lstm")
class LstmSeq2SeqEncoder(PytorchSeq2SeqWrapper):
    """
    Registered as a `Seq2SeqEncoder` with name "lstm".
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        num_layers: int = 1,
        bias: bool = True,
        dropout: float = 0.0,
        bidirectional: bool = False,
    ):
        module = torch.nn.LSTM(
            input_size=input_size,
            hidden_size=hidden_size,
            num_layers=num_layers,
            bias=bias,
            batch_first=True,
            dropout=dropout,
            bidirectional=bidirectional,
        )
        super().__init__(module=module)


@Seq2SeqEncoder.register("rnn")
class RnnSeq2SeqEncoder(PytorchSeq2SeqWrapper):
    """
    Registered as a `Seq2SeqEncoder` with name "rnn".
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        num_layers: int = 1,
        nonlinearity: str = "tanh",
        bias: bool = True,
        dropout: float = 0.0,
        bidirectional: bool = False,
    ):
        module = torch.nn.RNN(
            input_size=input_size,
            hidden_size=hidden_size,
            num_layers=num_layers,
            nonlinearity=nonlinearity,
            bias=bias,
            batch_first=True,
            dropout=dropout,
            bidirectional=bidirectional,
        )
        super().__init__(module=module)


class PytorchSeq2SeqWrapper_MLP(Seq2SeqEncoder):
    def __init__(self, module: torch.nn.Module) -> None:
        try:
            if not module.batch_first:
                raise ValueError("PytorchSeq2SeqWrapper only supports batch_first=True")
        except AttributeError:
            pass
        super().__init__()
        self._module = module
        try:
            self._is_bidirectional = cast(bool, self._module.bidirectional)
        except AttributeError:
            self._is_bidirectional = False
        if self._is_bidirectional:
            self._num_directions = 2
        else:
            self._num_directions = 1

    def get_input_dim(self) -> int:
        return cast(int, self._module.in_features)  # 修正

    def get_output_dim(self) -> int:
        return cast(int, self._module.out_features) * self._num_directions  # 修正

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        output = self._module(inputs)
        return cast(torch.Tensor, output)
    

@Seq2SeqEncoder.register("mlp")
class MlpSeq2SeqEncoder(PytorchSeq2SeqWrapper_MLP):
    """
    Registered as a `Seq2SeqEncoder` with name "mlp".
    """

    def __init__(
            self,
            input_size: int,
            hidden_size: int,
            bias: bool = True
        ):
        module = nn.Linear(in_features=input_size, out_features=hidden_size, bias=bias)
        super().__init__(module=module)